/*-
 * Copyright (c) 1990-1991 The Regents of the University of California.
 * All rights reserved.
 *
 * This code is derived from the Stanford/CMU enet packet filter,
 * (net/enet.c) distributed as part of 4.3BSD, and code contributed
 * to Berkeley by Steven McCanne and Van Jacobson both of Lawrence
 * Berkeley Laboratory.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 *	@(#)bpf.c	7.5 (Berkeley) 7/15/91
 *
 * static char rcsid[] =
 * "$Header: /home/adachi/UNIX/NeXT/bpf_NeXT/bpf_NeXT-current/RCS/bpf.c,v 0.3 1995/03/28 18:34:46 adachi Exp adachi $";
 */

#ifdef NeXT
/* Modified by S.Adachi for NEXTSTEP. */
static char rcsid[] = "$Id: bpf.c,v 0.3 1995/03/28 18:34:46 adachi Exp adachi $";
#endif

#include "bpfilter.h"

#if NBPFILTER > 0

#ifndef __GNUC__
#define inline
#else
#define inline __inline__
#endif

#include <sys/param.h>
#include <sys/systm.h>
#include <sys/mbuf.h>
#include <sys/buf.h>
#include <sys/dir.h> 
#include <sys/time.h>
/* NEXTSTEP does not use the U struct for the select system call. */
#ifndef NeXT
#include <sys/proc.h>
#include <sys/user.h>
#endif /* NOT NeXT */
#include <sys/ioctl.h>
#ifdef NeXT
#include <sys/printf.h>
#include <dev/ldd.h>
#include <kernserv/prototypes.h>
/* #include <machine/spl.h> */
#ifdef m68k
#include <kernserv/m68k/spl.h>
#else /* Not m68k */
#ifdef i386
#include <kernserv/i386/spl.h>
#else /* Not i386 */
NEXTSTEP seems to be ported to a new platform othe than m68k and i386...;
#endif /* Not i386 */
#endif /* Not m68k */

#else /* NeXT */
#include <sys/map.h>
#endif/* NeXT */

#include <sys/file.h>
#if defined(sparc) && BSD < 199103
#include <sys/stream.h>
#endif
#include <sys/tty.h>
#include <sys/uio.h>

#include <sys/protosw.h>
#include <sys/socket.h>
#include <net/if.h>

#include <net/bpf.h>
#include <net/bpfdesc.h>

#include <sys/errno.h>

#include <netinet/in.h>
#include <netinet/if_ether.h>
#include <sys/kernel.h>

#ifdef NeXT
#include <sys/conf.h>
#include <syslog.h>

#define LOG_DEBUG2 (LOG_DEBUG + 1)
#define LOG_DEBUG3 (LOG_DEBUG + 2)

/* kernel functions which has no descritions in header files. */
extern int uiomove(caddr_t addr, unsigned int count, enum uio_rw direction,
		   struct uio *uio);
extern int m_freem(register struct mbuf *m);
extern int selthreadcache(void **waiterPtr);
extern void selthreadclear(void **waiterPtr);
extern int selwakeup(void *waiter, int collided);

/* functions defined in other source files. */
extern int ifpromisc(struct ifnet *ifp, int pswitch);
extern int ifmulticast(struct ifnet *ifp, struct ifreq *ifr, int mswitch);
extern int bpf_validate(struct bpf_insn *f, int len);
extern void bpf_next_log(int log_level, char *msg,
			 int arg1, int arg2, int arg3, int arg4, int arg5);

/* verbose level */
extern int bpf_verbose;

/* It is better to include the following forward declarations
   for all platforms. */
int bpf_setf(struct bpf_d *d, struct bpf_program *fp);
int bpf_select();
#endif /* NeXT */

/*
 * Older BSDs don't have kernel malloc.
 */
#if BSD < 199103
#ifndef NeXT
extern bcopy();
#endif
static caddr_t bpf_alloc();
#include <net/bpf_compat.h>
#define BPF_BUFSIZE (MCLBYTES-8)
#define UIOMOVE(cp, len, code, uio) uiomove(cp, len, code, uio)
#else
#define BPF_BUFSIZE 4096
#define UIOMOVE(cp, len, code, uio) uiomove(cp, len, uio)
#endif

#define PRINET  26			/* interruptible */

/*
 * The default read buffer size is patchable.
 */
int bpf_bufsize = BPF_BUFSIZE;

/* ***** Added by S.Adachi ***** */
#ifdef RING_BUFFER
#ifndef BPF_NBUF
#define BPF_NBUF 8
#endif
int bpf_nbuf = BPF_NBUF;
#endif

/*
 *  bpf_iflist is the list of interfaces; each corresponds to an ifnet
 *  bpf_dtab holds the descriptors, indexed by minor device #
 */
struct bpf_if	*bpf_iflist;
struct bpf_d	bpf_dtab[NBPFILTER];

static void	bpf_ifname();
static void	catchpacket();
static void	bpf_freed();
static int	bpf_setif();
#ifndef NeXT
static int	bpf_initd();
#endif /* NeXT */
static int	bpf_allocbufs();

static int
bpf_movein(uio, linktype, mp, sockp)
	register struct uio *uio;
	int linktype;
	register struct mbuf **mp;
	register struct sockaddr *sockp;
{
	struct mbuf *m;
	int error;
	int len;
	int hlen;

	/*
	 * Build a sockaddr based on the data link layer type.
	 * We do this at this level because the ethernet header
	 * is copied directly into the data field of the sockaddr.
	 * In the case of SLIP, there is no header and the packet
	 * is forwarded as is.
	 * Also, we are careful to leave room at the front of the mbuf
	 * for the link level header.
	 */
	switch (linktype) {

	case DLT_SLIP:
		sockp->sa_family = AF_INET;
		hlen = 0;
		break;

	case DLT_EN10MB:
		sockp->sa_family = AF_UNSPEC;
		/* XXX Would MAXLINKHDR be better? */
		hlen = sizeof(struct ether_header);
		break;

	case DLT_FDDI:
		sockp->sa_family = AF_UNSPEC;
		/* XXX 4(FORMAC)+6(dst)+6(src)+3(LLC)+5(SNAP) */
		hlen = 24;
		break;

	case DLT_NULL:
		sockp->sa_family = AF_UNSPEC;
		hlen = 0;
		break;

	default:
		return (EIO);
	}

	len = uio->uio_resid;
	if ((unsigned)len > MCLBYTES)
		return (EIO);

	MGET(m, M_WAIT, MT_DATA);
	if (m == 0)
		return (ENOBUFS);
	if (len > MLEN) {
#if BSD >= 199103
		MCLGET(m, M_WAIT);
		if ((m->m_flags & M_EXT) == 0) {
#else
		MCLGET(m);
		if (m->m_len != MCLBYTES) {
#endif
			error = ENOBUFS;
			goto bad;
		}
	}
	m->m_len = len;
	*mp = m;
	/*
	 * Make room for link header.
	 */
	if (hlen != 0) {
		m->m_len -= hlen;
#if BSD >= 199103
		m->m_data += hlen; /* XXX */
#else
		m->m_off += hlen;
#endif
		error = UIOMOVE((caddr_t)sockp->sa_data, hlen, UIO_WRITE, uio);
		if (error)
			goto bad;
	}
	error = UIOMOVE(mtod(m, caddr_t), len - hlen, UIO_WRITE, uio);
	if (!error)
		return (0);
 bad:
	m_freem(m);
	return (error);
}

/*
 * Attach file to the bpf interface, i.e. make d listen on bp.
 * Must be called at splimp.
 */
static void
bpf_attachd(d, bp)
	struct bpf_d *d;
	struct bpf_if *bp;
{
	/*
	 * Point d at bp, and add d to the interface's list of listeners.
	 * Finally, point the driver's bpf cookie at the interface so
	 * it will divert packets to bpf.
	 */
	d->bd_bif = bp;
	d->bd_next = bp->bif_dlist;
	bp->bif_dlist = d;

	*bp->bif_driverp = bp;
}

/*
 * Detach a file from its interface.
 */
static void
bpf_detachd(d)
	struct bpf_d *d;
{
	struct bpf_d **p;
	struct bpf_if *bp;

#ifdef NeXT
	bpf_next_log(LOG_DEBUG, "bpf_detached entered (d = 0x%x)\n",
		     (int)d, 0, 0, 0, 0);
#endif

	bp = d->bd_bif;
	/*
	 * Check if this descriptor had requested promiscuous mode.
	 * If so, turn it off.
	 */
	if (d->bd_promisc) {
		d->bd_promisc = 0;
		if (ifpromisc(bp->bif_ifp, 0))
			/*
			 * Something is really wrong if we were able to put
			 * the driver into promiscuous mode, but can't
			 * take it out.
			 */
			panic("bpf: ifpromisc failed");
	}
	/* Remove d from the interface's descriptor list. */
	p = &bp->bif_dlist;
	while (*p != d) {
		p = &(*p)->bd_next;
		if (*p == 0)
			panic("bpf_detachd: descriptor not in list");
	}
	*p = (*p)->bd_next;
	if (bp->bif_dlist == 0)
		/*
		 * Let the driver know that there are no more listeners.
		 */
		*d->bd_bif->bif_driverp = 0;
	d->bd_bif = 0;
#ifdef NeXT
	bpf_next_log(LOG_DEBUG, "bpf_detached leaving ... (d = 0x%x)\n",
		     (int)d, 0, 0, 0, 0);
#endif
}


/*
 * Mark a descriptor free by making it point to itself.
 * This is probably cheaper than marking with a constant since
 * the address should be in a register anyway.
 */
#define D_ISFREE(d) ((d) == (d)->bd_next)
#define D_MARKFREE(d) ((d)->bd_next = (d))
#define D_MARKUSED(d) ((d)->bd_next = 0)

/*
 * Open ethernet device.  Returns ENXIO for illegal minor device number,
 * EBUSY if file is open by another process.
 */
/* ARGSUSED */
int
bpfopen(dev, flag)
	dev_t dev;
	int flag;
{
	register struct bpf_d *d;
	int error;

#ifdef NeXT
	bpf_next_log(LOG_DEBUG, "bpfopen entered\n", 0, 0, 0, 0, 0);
#endif

	if (minor(dev) >= NBPFILTER) {
	  error = ENXIO;
	  goto END;
	}

	/*
	 * Each minor can be opened by only one process.  If the requested
	 * minor is in use, return EBUSY.
	 */
	d = &bpf_dtab[minor(dev)];
	if (!D_ISFREE(d)) {
	  error = EBUSY;
	  goto END;
	}

	/* Mark "free" and do most initialization. */
	bzero((char *)d, sizeof(*d));
	d->bd_bufsize = bpf_bufsize;
#ifdef RING_BUFFER
	d->bd_nbuf = bpf_nbuf;
#endif /* RING_BUFFER */

	error = 0;
      END:
#ifdef NeXT
	bpf_next_log(LOG_DEBUG, "bpfopen leaving ...(error = %d)\n",
		     error, 0, 0, 0, 0);
#endif
	return error;
}

/*
 * Close the descriptor by detaching it from its interface,
 * deallocating its buffers, and marking it free.
 */
/* ARGSUSED */
int
bpfclose(dev, flag)
	dev_t dev;
	int flag;
{
	register struct bpf_d *d = &bpf_dtab[minor(dev)];
	register int s;

#ifdef NeXT
	bpf_next_log(LOG_DEBUG, "bpfclose entered\n", 0, 0, 0, 0, 0);
#endif

#ifdef NeXT
	if (d->bd_selthread)
	  selthreadclear(&d->bd_selthread);
#endif /* NeXT */

	s = splimp();
	if (d->bd_bif)
		bpf_detachd(d);
	splx(s);
	bpf_freed(d);

#ifdef NeXT
	bpf_next_log(LOG_DEBUG, "bpfclose leaving ...\n", 0, 0, 0, 0, 0);
#endif
	return (0);
}

/*
 * Support for SunOS, which does not have tsleep.
 */
#if BSD < 199103
static int
bpf_timeout(arg)
	caddr_t arg;
{
	struct bpf_d *d = (struct bpf_d *)arg;
	d->bd_timedout = 1;
	wakeup(arg);
#ifdef NeXT
	return 0;
#endif
}

#define BPF_SLEEP(chan, pri, s, t) bpf_sleep((struct bpf_d *)chan)

int
bpf_sleep(d)
	register struct bpf_d *d;
{
	register int rto = d->bd_rtout;
	register int st;

#ifdef NeXT
	bpf_next_log(LOG_DEBUG, "bpf_sleep: entered (d = 0x%x)\n",
		     (int)d, 0, 0, 0, 0);
#endif

	if (rto != 0) {
		d->bd_timedout = 0;
		timeout(bpf_timeout, (caddr_t)d, rto);
	}
	st = sleep((caddr_t)d, PRINET|PCATCH);
	if (rto != 0) {
		if (d->bd_timedout == 0)
			untimeout(bpf_timeout, (caddr_t)d);
		else if (st == 0)
			return EWOULDBLOCK;
	}
#ifdef NeXT
	bpf_next_log(LOG_DEBUG, "bpf_sleep: leaving (d = 0x%x)\n",
		     (int)d, 0, 0, 0, 0);
#endif
	return (st != 0) ? EINTR : 0;
}
#else
#define BPF_SLEEP tsleep
#endif

#ifdef RING_BUFFER
#define ROTATE_BUFFERS(d) \
        if ((d)->bd_sidx == (d)->bd_nbuf - 1) \
          (d)->bd_sidx = 0; \
	else \
          (d)->bd_sidx ++; \
        (d)->bd_len[(d)->bd_sidx] = 0; \
        (d)->bd_hcount ++;
#else /* Not RING_BUFFER */
/*
 * Rotate the packet buffers in descriptor d.  Move the store buffer
 * into the hold slot, and the free buffer into the store slot.
 * Zero the length of the new store buffer.
 */
#define ROTATE_BUFFERS(d) \
	(d)->bd_hbuf = (d)->bd_sbuf; \
	(d)->bd_hlen = (d)->bd_slen; \
	(d)->bd_sbuf = (d)->bd_fbuf; \
	(d)->bd_slen = 0; \
	(d)->bd_fbuf = 0;
#endif /* Not RING_BUFFER */

/*
 *  bpfread - read next chunk of packets from buffers
 */
int
bpfread(dev, uio)
	dev_t dev;
	register struct uio *uio;
{
	register struct bpf_d *d = &bpf_dtab[minor(dev)];
	int error;
	int s;

#ifdef NeXT
	bpf_next_log(LOG_DEBUG, "bpfread: entered (d = 0x%x)\n",
		     (int)d, 0, 0, 0, 0);
#endif

	/*
	 * Restrict application to use a buffer the same size as
	 * as kernel buffers.
	 */
	if (uio->uio_resid != d->bd_bufsize) {
	  error = EINVAL;
	  goto END;
	}

	s = splimp();
#ifdef RING_BUFFER
	/*
	 * If buffers are empty, then do a timed aleep, which
	 * ends when the timeout expires or when enough packets
	 * have arrived to fill a buffer.
	 */
	while (d->bd_hcount == 0) {
	  if (d->bd_immediate && d->bd_len[d->bd_sidx] > 0) {
	    /*
	     * A packet(s) either arrived since the previous
	     * read or arrived while we were asleep.
	     * Rotate the buffers and return what's here.
	     */
	    ROTATE_BUFFERS(d);
	    break;
	  }
	  error = BPF_SLEEP((caddr_t)d, PRINET|PCATCH, "bpf\n",
			    d->bd_rtout);
	  if (error == EINTR || error == ERESTART) {
	    splx(s);
	    goto END;
	  }
	  if (error == EWOULDBLOCK) {
	    /*
	     * On a timeout, return what's in the buffer,
	     * which may be nothing.  If there is something
	     * in the store buffer, we can rotate the buffers.
	     */
	    if (d->bd_hcount > 0)
	      /*
	       * We filled up the buffer in between
	       * getting the timeout and arriving
	       * here, so we don't need to rotate.
	       */
	      break;

	    if (d->bd_len[d->bd_sidx] == 0) {
	      splx(s);
	      error = 0;
	      goto END;
	    }
	    ROTATE_BUFFERS(d);
	    break;
	  }
	}
        /*
	 * At this point, we know we have something in the hold slot.
	 */
	splx(s);

	/*
	 * Move data from hold buffer into user space.
	 * We know the entire buffer is transferred since
	 * we checked above that the read buffer is bpf_bufsize bytes.
	 */
	error = UIOMOVE(d->bd_buf[d->bd_hidx], d->bd_len[d->bd_hidx],
			UIO_READ, uio);

	s = splimp();
	d->bd_len[d->bd_hidx] = 0;
	if (d->bd_hidx == d->bd_nbuf - 1)
	  d->bd_hidx = 0;
	else
	  d->bd_hidx ++;
	d->bd_hcount --;
	splx(s);

#else /* Not RING_BUFFER */
	/*
	 * If the hold buffer is empty, then do a timed sleep, which
	 * ends when the timeout expires or when enough packets
	 * have arrived to fill the store buffer.
	 */
	while (d->bd_hbuf == 0) {
		if (d->bd_immediate && d->bd_slen != 0) {
			/*
			 * A packet(s) either arrived since the previous
			 * read or arrived while we were asleep.
			 * Rotate the buffers and return what's here.
			 */
			ROTATE_BUFFERS(d);
			break;
		}
		error = BPF_SLEEP((caddr_t)d, PRINET|PCATCH, "bpf\n",
				  d->bd_rtout);
		if (error == EINTR || error == ERESTART) {
			splx(s);
			goto END;
		}
		if (error == EWOULDBLOCK) {
			/*
			 * On a timeout, return what's in the buffer,
			 * which may be nothing.  If there is something
			 * in the store buffer, we can rotate the buffers.
			 */
			if (d->bd_hbuf)
				/*
				 * We filled up the buffer in between
				 * getting the timeout and arriving
				 * here, so we don't need to rotate.
				 */
				break;

			if (d->bd_slen == 0) {
				splx(s);
				error = 0;
				goto END;
			}
			ROTATE_BUFFERS(d);
			break;
		}
	}
	/*
	 * At this point, we know we have something in the hold slot.
	 */
	splx(s);

	/*
	 * Move data from hold buffer into user space.
	 * We know the entire buffer is transferred since
	 * we checked above that the read buffer is bpf_bufsize bytes.
	 */
	error = UIOMOVE(d->bd_hbuf, d->bd_hlen, UIO_READ, uio);

	s = splimp();
	d->bd_fbuf = d->bd_hbuf;
	d->bd_hbuf = 0;
	d->bd_hlen = 0;
	splx(s);
#endif /* Not RING_BUFFER */

      END:

#ifdef NeXT
	bpf_next_log(LOG_DEBUG, "bpfread leaving... (error = %d)\n",
		     error, 0, 0, 0, 0);
#endif
	return (error);
}


/*
 * If there are processes sleeping on this descriptor, wake them up.
 */
static inline void
bpf_wakeup(d)
	register struct bpf_d *d;
{
#ifdef NeXT
        int s; /* @@ Necessary ? S.A. */
        bpf_next_log(LOG_DEBUG, "bpf_wakeup: entered (d = 0x%x)\n",
	       (int)d, 0, 0, 0, 0);
#endif
	wakeup((caddr_t)d);
#ifdef NeXT
        s = splimp(); /* @@ Necessary ? S.A. */
	if (d->bd_selthread) {
	  bpf_next_log(LOG_DEBUG, "bpf_wakeup: wakeup thread (0x%x)\n",
		       (int)d->bd_selthread, 0, 0, 0, 0);
	  selwakeup(d->bd_selthread, (int)d->bd_selcoll);
	  selthreadclear(&d->bd_selthread);
	  d->bd_selcoll = 0;
	  d->bd_selthread = 0;
	}
        splx(s);  /* @@ Necessary ? S.A. */
#else /* NeXT */
#if BSD >= 199103
	selwakeup(&d->bd_sel);
	/* XXX */
	d->bd_sel.si_pid = 0;
#else
	if (d->bd_selproc) {
		selwakeup(d->bd_selproc, (int)d->bd_selcoll);
		d->bd_selcoll = 0;
		d->bd_selproc = 0;
	}
#endif
#endif /* NeXT */
#ifdef NeXT
	  bpf_next_log(LOG_DEBUG, "bpf_wakeup: leaving ... (d = 0x%x)\n",
		       (int)d, 0, 0, 0, 0);
#endif
}

int
bpfwrite(dev, uio)
	dev_t dev;
	struct uio *uio;
{
	register struct bpf_d *d = &bpf_dtab[minor(dev)];
	struct ifnet *ifp;
	struct mbuf *m;
	int error, s;
	static struct sockaddr dst;

#ifdef NeXT
	bpf_next_log(LOG_DEBUG, "bpfwrite: entered (d = 0x%x)\n",
		     (int)d, 0, 0, 0, 0);
#endif

	if (d->bd_bif == 0) {
	  error = ENXIO;
	  goto END;
	}

	ifp = d->bd_bif->bif_ifp;

	if (uio->uio_resid == 0) {
	  error = 0;
	  goto END;
	}
	if (uio->uio_resid > ifp->if_mtu) {
	  error = EMSGSIZE;
	  goto END;
	}

	error = bpf_movein(uio, (int)d->bd_bif->bif_dlt, &m, &dst);
	if (error)
	  goto END;

#ifdef NeXT
	{
	  struct ether_header eh;
	  struct in_addr idst;
	  int error, size;
	  netbuf_t nb;

	  /* Set the Ethernet header according to the specified
	     address family. */
	  switch (dst.sa_family) {
	  case AF_UNSPEC:
	    bcopy(&dst.sa_data, &eh, sizeof(eh));
	    break;
	  case AF_INET:
	    idst = ((struct sockaddr_in *)&dst)->sin_addr;
	    /*
	     *   Resolve the en address using arp. Return 0 if the address
	     *   wasn't resolved.
	     *
	     *   XXX: trailers not supported for output.
	     */
#if 0
	    /* %%%%% NOT YET IMPLEMENTED %%%%% */
	    /* ????? */
	    eh.ether_type = htons(ETHERTYPE_IP);
	    break;
#else
	    bpf_next_log(LOG_WARNING,
			 "bpfwrite: Sorry, AF_INET is not supported yet.\n",
			 0, 0, 0, 0, 0);
	    m_free(m);
	    error = EAFNOSUPPORT;
	    goto END;
#endif
	  default:
	    m_free(m);
	    error = EAFNOSUPPORT;
	    goto END;
	  }

	  /* Get the netbuf for the Ethernet interface. */
	  nb = if_getbuf((netif_t)ifp);
	  if (nb == NULL) {
	    m_free(m);
	    error = ENOBUFS;
	    goto END;
	  }

	  /* Calculate the size of the portion of the netbuf "nb"
	     which will be actually used. */
	  size = ETHERHDRSIZE +  m->m_len;

	  /* Check that the netbuf "nb" is large enough. */
	  if (size > nb_size(nb)) {
	    error = EMSGSIZE;
	    goto END;
	  }

	  /* Write the Ethernet header and data. */
	  nb_write(nb, /* ETYPEOFFSET */ 6 + 6, sizeof(eh.ether_type),
		   (void *)&eh.ether_type);
	  nb_write(nb, ETHERHDRSIZE, m->m_len, mtod(m, caddr_t));

	  /* Trim the tail of the netbuf to match the actual size. */
	  nb_shrink_bot(nb, nb_size(nb) - size);

	  /* Request the Ethernet driver to send the packet. */
	  s = splnet();
	  error = if_output((netif_t)ifp, nb, (void *)&eh.ether_dhost);
	  splx(s);

	  /* Free the mbuf. */
	  m_free(m);
	}
#else /* Not NeXT */
	s = splnet();
#if BSD >= 199103
	error = (*ifp->if_output)(ifp, m, &dst, (struct rtentry *)0);
#else
	error = (*ifp->if_output)(ifp, m, &dst);
#endif
	splx(s);
#endif /* Not NeXT */
	/*
	 * The driver frees the mbuf.
	 */
      END:
#ifdef NeXT
	bpf_next_log(LOG_DEBUG, "bpwrite leaving...(error = %d)\n",
		     error, 0, 0, 0, 0);
#endif
	return (error);
}

/*
 * Reset a descriptor by flushing its packet buffer and clearing the
 * receive and drop counts.  Should be called at splimp.
 */
static void
reset_d(d)
     struct bpf_d *d;
{
#ifdef NeXT
  bpf_next_log(LOG_DEBUG, "reset_d entered (d = 0x%x)\n",
	       (int)d, 0, 0, 0, 0);
#endif
#ifdef RING_BUFFER
  if (d->bd_len) {
    int i;
    for (i = 0; i < d->bd_nbuf; i++)
      d->bd_len[i] = 0;
  }
  d->bd_sidx = 0;
  d->bd_hidx = 0;
  d->bd_hcount = 0;
#else /* Not RING_BUFFER */
  if (d->bd_hbuf) {
    /* Free the hold buffer. */
    d->bd_fbuf = d->bd_hbuf;
    d->bd_hbuf = 0;
  }
  d->bd_slen = 0;
  d->bd_hlen = 0;
#endif /* Not RING_BUFFER */
  d->bd_rcount = 0;
  d->bd_dcount = 0;
#ifdef NeXT
  bpf_next_log(LOG_DEBUG, "reset_d leaving ... (d = 0x%x)\n",
	       (int)d, 0, 0, 0, 0);
#endif
}

/*
 *  FIONREAD		Check for read packet available.
 *  SIOCGIFADDR		Get interface address - convenient hook to driver.
 *  BIOCGBLEN		Get buffer len [for read()].
 *  BIOCSETF		Set ethernet read filter.
 *  BIOCFLUSH		Flush read packet buffer.
 *  BIOCPROMISC		Put interface into promiscuous mode.
 *  BIOCGDLT		Get link layer type.
 *  BIOCGETIF		Get interface name.
 *  BIOCSETIF		Set interface.
 *  BIOCSRTIMEOUT	Set read timeout.
 *  BIOCGRTIMEOUT	Get read timeout.
 *  BIOCGSTATS		Get packet stats.
 *  BIOCIMMEDIATE	Set immediate mode.
 *  BIOCVERSION		Get filter language version.
 */
#ifdef NeXT
/*
 *  The following two controls are added for NEXTSTEP 3.2J,
 *  since it has not yet supported officially multicast,
 *  though the kernel actually seems to have the essential part to support
 *  the multicast. The multicast is necessary for CAP configured for
 *  AppleTalk Phase 2.
 *
 *  If it is possible, the muticast should be controled by ioctl calls
 *  applied to a socket, which is opend by the user. At the present time,
 *  this is impossible.
 *
 *  Please enhance the next release of NEXTSTEP
 *  to support officially the multicast, NeXT Inc.!
 *
 *  SIOCADDMULTI        Add an multicast address.
 *  SIOCDELMULTI        Delete an multicast address.
 */
#endif /* NeXT */
/* ARGSUSED */
int
bpfioctl(dev, cmd, addr, flag)
	dev_t dev;
	int cmd;
	caddr_t addr;
	int flag;
{
	register struct bpf_d *d = &bpf_dtab[minor(dev)];
	int s, error = 0;

#ifdef NeXT
	bpf_next_log(LOG_DEBUG, "bpfioctl entered (d = 0x%x)\n",
		     (int)d, 0, 0, 0, 0);
#endif

	switch (cmd) {

	default:
		error = EINVAL;
		break;

	/*
	 * Check for read packet available.
	 */
	case FIONREAD:
		{
			int n;

#ifdef NeXT
			bpf_next_log(LOG_DEBUG, "bpfioctl: FIONREAD\n",
				     0, 0, 0, 0, 0);
#endif
			s = splimp();
#ifdef RING_BUFFER
			{
			  int i, j = d->bd_hidx;
			  n = d->bd_len[d->bd_sidx];
			  for (i = 0; i < d->bd_hcount; i++) {
			    n += d->bd_len[j];
			    if (++j == d->bd_nbuf)
			      j = 0;
			  }
			}
#else /* Not RING_BUFFER */
			n = d->bd_slen;
			if (d->bd_hbuf)
				n += d->bd_hlen;
#endif /* Not RING_BUFFER */
			splx(s);

			*(int *)addr = n;
			break;
		}

	case SIOCGIFADDR:
		{
			struct ifnet *ifp;

#ifdef NeXT
			bpf_next_log(LOG_DEBUG, "bpfioctl: SIOCGIFADDR\n",
				     0, 0, 0, 0, 0);
#endif
			if (d->bd_bif == 0) {
				error = EINVAL;
#ifdef NeXT
				bpf_next_log(LOG_DEBUG, "bpfioctl: NO bd_bif\n",
					     0, 0, 0, 0, 0);
#endif
			}
			else {
				ifp = d->bd_bif->bif_ifp;
#ifdef NeXT
				{
				  /* @@ Dirty Hack code: need investigation. */
#if 0
				  /* This code does not work... Why ?
				     S.Aadachi (94/07/17) */
				  if_ioctl_t pack;
				  pack.ioctl_command = cmd;
				  pack.ioctl_data = addr;
				  error = (*ifp->if_control)(ifp,
							     IFCONTROL_UNIXIOCTL,
							     &pack);
#else
				  /* This code is OK. S.Aadachi (94/07/17) */
				  /* HACK HACK HACK ... */
				  /* struct sockaddr
				       {u_short sa_family; char sa_data[14]; };
				     struct inreq {
				       char ifr_name[IFNAMSIZ];
				       union {
				         struct sockaddr ifru_addr;
					 ...
				       } ifr_ifru;
				     } * addr; */
				  struct sockaddr *sa
				    = (struct sockaddr *)(addr + IFNAMSIZ);
				  sa->sa_family = AF_DLI;/* Direct data link */
				  error = (*ifp->if_control)(ifp,
							     IFCONTROL_GETADDR,
							     &sa->sa_data);
				  /* sa_data[0],...,sa_data[5] were filled. */
				  bzero(sa->sa_data + 6, 8);
#endif
				  if (error != 0)
				    bpf_next_log(LOG_DEBUG,
						 "bpfioctl: ioctl for lower if failed\n",
						 0, 0, 0, 0, 0);
				}
#else
				error = (*ifp->if_ioctl)(ifp, cmd, addr);
#endif
			}
			break;
		}

	/*
	 * Get buffer len [for read()].
	 */
	case BIOCGBLEN:
#ifdef NeXT
		bpf_next_log(LOG_DEBUG, "bpfioctl: BIOCGBLEN\n", 0, 0, 0, 0, 0);
#endif
		*(u_int *)addr = d->bd_bufsize;
		break;

	/*
	 * Set buffer length.
	 */
	case BIOCSBLEN:
#ifdef NeXT
		bpf_next_log(LOG_DEBUG, "bpfioctl: BIOCSBLEN\n", 0, 0, 0, 0, 0);
#endif
#if BSD < 199103
		error = EINVAL;
#else
		if (d->bd_bif != 0)
			error = EINVAL;
		else {
			register u_int size = *(u_int *)addr;

			if (size > BPF_MAXBUFSIZE)
				*(u_int *)addr = size = BPF_MAXBUFSIZE;
			else if (size < BPF_MINBUFSIZE)
				*(u_int *)addr = size = BPF_MINBUFSIZE;
			d->bd_bufsize = size;
		}
#endif
		break;

	/*
	 * Set link layer read filter.
	 */
	case BIOCSETF:
#ifdef NeXT
		bpf_next_log(LOG_DEBUG, "bpfioctl: BIOCSETF\n", 0, 0, 0, 0, 0);
#endif
		error = bpf_setf(d, (struct bpf_program *)addr);
		break;

	/*
	 * Flush read packet buffer.
	 */
	case BIOCFLUSH:
#ifdef NeXT
		bpf_next_log(LOG_DEBUG, "bpfioctl: BIOCFLUSH\n", 0, 0, 0, 0, 0);
#endif
		s = splimp();
		reset_d(d);
		splx(s);
		break;

	/*
	 * Put interface into promiscuous mode.
	 */
	case BIOCPROMISC:
#ifdef NeXT
		bpf_next_log(LOG_DEBUG, "bpfioctl: BIOCPROMISC\n",
			     0, 0, 0, 0, 0);
#endif
		if (d->bd_bif == 0) {
			/*
			 * No interface attached yet.
			 */
			error = EINVAL;
			break;
		}
		s = splimp();
		if (d->bd_promisc == 0) {
			error = ifpromisc(d->bd_bif->bif_ifp, 1);
			if (error == 0)
				d->bd_promisc = 1;
		}
		splx(s);
		break;

	/*
	 * Get device parameters.
	 */
	case BIOCGDLT:
#ifdef NeXT
		bpf_next_log(LOG_DEBUG, "bpfioctl: BIOCGDLT\n", 0, 0, 0, 0, 0);
#endif
		if (d->bd_bif == 0)
			error = EINVAL;
		else
			*(u_int *)addr = d->bd_bif->bif_dlt;
		break;

	/*
	 * Set interface name.
	 */
	case BIOCGETIF:
#ifdef NeXT
		bpf_next_log(LOG_DEBUG, "bpfioctl: BIOCGETIF\n", 0, 0, 0, 0, 0);
#endif
		if (d->bd_bif == 0)
			error = EINVAL;
		else
			bpf_ifname(d->bd_bif->bif_ifp, (struct ifreq *)addr);
		break;

	/*
	 * Set interface.
	 */
	case BIOCSETIF:
#ifdef NeXT
		bpf_next_log(LOG_DEBUG, "bpfioctl: BIOCSETIF\n", 0, 0, 0, 0, 0);
#endif
		error = bpf_setif(d, (struct ifreq *)addr);
		break;

	/*
	 * Set read timeout.
	 */
	case BIOCSRTIMEOUT:
		{
			struct timeval *tv = (struct timeval *)addr;
			u_long msec;

#ifdef NeXT
			bpf_next_log(LOG_DEBUG, "bpfioctl: BIOCSRTIMEOUT\n",
				     0, 0, 0, 0, 0);
#endif
			/* Compute number of milliseconds. */
			msec = tv->tv_sec * 1000 + tv->tv_usec / 1000;
			/* Scale milliseconds to ticks.  Assume hard
			   clock has millisecond or greater resolution
			   (i.e. tick >= 1000).  For 10ms hardclock,
			   tick/1000 = 10, so rtout<-msec/10. */
			d->bd_rtout = msec / (tick / 1000);
			break;
		}

	/*
	 * Get read timeout.
	 */
	case BIOCGRTIMEOUT:
		{
			struct timeval *tv = (struct timeval *)addr;
			u_long msec = d->bd_rtout;

#ifdef NeXT
			bpf_next_log(LOG_DEBUG, "bpfioctl: BIOCGRTIMEOUT\n",
				     0, 0, 0, 0, 0);
#endif
			msec *= tick / 1000;
			tv->tv_sec = msec / 1000;
			tv->tv_usec = msec % 1000;
			break;
		}

	/*
	 * Get packet stats.
	 */
	case BIOCGSTATS:
		{
			struct bpf_stat *bs = (struct bpf_stat *)addr;

#ifdef NeXT
			bpf_next_log(LOG_DEBUG, "bpfioctl: BIOCGSTATS\n",
				     0, 0, 0, 0, 0);
#endif
			bs->bs_recv = d->bd_rcount;
			bs->bs_drop = d->bd_dcount;
			break;
		}

	/*
	 * Set immediate mode.
	 */
	case BIOCIMMEDIATE:
#ifdef NeXT
		bpf_next_log(LOG_DEBUG, "bpfioctl: BIOCIMMEDIATE\n",
			     0, 0, 0, 0, 0);
#endif
		d->bd_immediate = *(u_int *)addr;
		break;

	case BIOCVERSION:
		{
			struct bpf_version *bv = (struct bpf_version *)addr;

#ifdef NeXT
			bpf_next_log(LOG_DEBUG, "bpfioctl: BIOCVERSION\n",
				     0, 0, 0, 0, 0);
#endif
			bv->bv_major = BPF_MAJOR_VERSION;
			bv->bv_minor = BPF_MINOR_VERSION;
			break;
		}
#ifdef NeXT
	/*
	 * Add/Delete a multicast address.
	 */
	case SIOCADDMULTI:
	case SIOCDELMULTI:
		{
		  int add = (cmd == SIOCADDMULTI);

		  if (add)
		    bpf_next_log(LOG_DEBUG, "bpfioctl: SIOCADDMULTI\n",
				 0, 0, 0, 0, 0);
		  else
		    bpf_next_log(LOG_DEBUG, "bpfioctl: SIOCDELMULTI\n",
				 0, 0, 0, 0, 0);
		  if (d->bd_bif == 0) {
		    /*
		     * No interface attached yet.
		     */
		    bpf_next_log(LOG_DEBUG, "bpfioctl: No interface attached yet.\n",
				 0, 0, 0, 0, 0);
		    error = EINVAL;
		    break;
		  }
		  s = splimp();
		  error = ifmulticast(d->bd_bif->bif_ifp,
				      (struct ifreq *)addr,
				      add);
		  splx(s);
		  break;
		}
#endif /* NeXT */
	}
#ifdef NeXT
	bpf_next_log(LOG_DEBUG, "bpfioctl leaving ... (d = 0x%x, error = %d)\n",
		     (int)d, error, 0, 0, 0);
#endif
	return (error);
}

/*
 * Set d's packet filter program to fp.  If this file already has a filter,
 * free it and replace it.  Returns EINVAL for bogus requests.
 */
int
bpf_setf(d, fp)
	struct bpf_d *d;
	struct bpf_program *fp;
{
	struct bpf_insn *fcode, *old;
	u_int flen, size;
	int s;
	int error;

#ifdef NeXT
	bpf_next_log(LOG_DEBUG, "bpf_setf enterd (d = 0x%x)\n",
		     (int)d, 0, 0, 0, 0);
#endif
	old = d->bd_filter;
	if (fp->bf_insns == 0) {
		if (fp->bf_len != 0) {
		  error = EINVAL;
		  goto END;
		}
		s = splimp();
		d->bd_filter = 0;
		reset_d(d);
		splx(s);
		if (old != 0)
			free((caddr_t)old, M_DEVBUF);
		error = 0;
		goto END;
	}
	flen = fp->bf_len;
	if (flen > BPF_MAXINSNS){
	  error = EINVAL;
	  goto END;
	}
	size = flen * sizeof(*fp->bf_insns);
	fcode = (struct bpf_insn *)malloc(size, M_DEVBUF, M_WAITOK);
	if (copyin((caddr_t)fp->bf_insns, (caddr_t)fcode, size) == 0 &&
	    bpf_validate(fcode, (int)flen)) {
		s = splimp();
		d->bd_filter = fcode;
		reset_d(d);
		splx(s);
		if (old != 0)
			free((caddr_t)old, M_DEVBUF);
		error = 0;
		goto END;
	}
	free((caddr_t)fcode, M_DEVBUF);
	error = EINVAL;
      END:
#ifdef NeXT
	bpf_next_log(LOG_DEBUG, "bpf_setf leaving ... (d = 0x%x, error = %d)\n",
		     (int)d, error, 0, 0, 0);
#endif
	return error;
}

/*
 * Detach a file from its current interface (if attached at all) and attach
 * to the interface indicated by the name stored in ifr.
 * Return an errno or 0.
 */
static int
bpf_setif(d, ifr)
	struct bpf_d *d;
	struct ifreq *ifr;
{
	struct bpf_if *bp;
	char *cp;
	int unit, s, error;

	/*
	 * Separate string into name part and unit number.  Put a null
	 * byte at the end of the name part, and compute the number.
	 * If the a unit number is unspecified, the default is 0,
	 * as initialized above.  XXX This should be common code.
	 */
	unit = 0;
	cp = ifr->ifr_name;
	cp[sizeof(ifr->ifr_name) - 1] = '\0';
	while (*cp++) {
		if (*cp >= '0' && *cp <= '9') {
			unit = *cp - '0';
			*cp++ = '\0';
			while (*cp)
				unit = 10 * unit + *cp++ - '0';
			break;
		}
	}
	/*
	 * Look through attached interfaces for the named one.
	 */
	for (bp = bpf_iflist; bp != 0; bp = bp->bif_next) {
		struct ifnet *ifp = bp->bif_ifp;

		if (ifp == 0 || unit != ifp->if_unit
		    || strcmp(ifp->if_name, ifr->ifr_name) != 0)
			continue;
		/*
		 * We found the requested interface.
		 * If it's not up, return an error.
		 * Allocate the packet buffers if we need to.
		 * If we're already attached to requested interface,
		 * just flush the buffer.
		 */
		/* *** Dirty and temporal hack for NEXTSTEP for Intel *** */
		/* The flag "if_flags" for Ethernet interface (class == "real")
		   always takes the value zero for NeXTSTEP 3.3J for Intel.
		   This is found by Suzuki Yoshiaki <suzuki@sowa.is.uec.ac.jp>
		   and Kunihiro Yamana <yaman@open.rd.nttdata.jp>.
		   I think that this is a bug in NEXTSTEP itself!
		   (S.Adachi, 96/02/20) */
#if !defined(NeXT) || !defined(i386)
		if ((ifp->if_flags & IFF_UP) == 0)
			return (ENETDOWN);
#endif /* defined(NeXT) && defined(i386) */
#ifdef RING_BUFFER
		if (d->bd_buf == 0) {
			error = bpf_allocbufs(d);
			if (error != 0)
				return (error);
		}
#else /* Not RING_BUFFER */
		if (d->bd_sbuf == 0) {
		        error = bpf_allocbufs(d);
			if (error != 0)
				return (error);
		}
#endif /* Not RING_BUFFER */
		s = splimp();
		if (bp != d->bd_bif) {
			if (d->bd_bif)
				/*
				 * Detach if attached to something else.
				 */
				bpf_detachd(d);

			bpf_attachd(d, bp);
		}
		reset_d(d);
		splx(s);
		return (0);
	}
	/* Not found. */
	return (ENXIO);
}

/*
 * Convert an interface name plus unit number of an ifp to a single
 * name which is returned in the ifr.
 */
static void
bpf_ifname(ifp, ifr)
	struct ifnet *ifp;
	struct ifreq *ifr;
{
	char *s = ifp->if_name;
	char *d = ifr->ifr_name;

	while (*d++ = *s++)
		continue;
	/* XXX Assume that unit number is less than 10. */
	*d++ = ifp->if_unit + '0';
	*d = '\0';
}

/*
 * The new select interface passes down the proc pointer; the old select
 * stubs had to grab it out of the user struct.  This glue allows either case.
 */
#if BSD >= 199103
#define bpf_select bpfselect
#else
int
bpfselect(dev, rw)
	register dev_t dev;
	int rw;
{
#ifdef NeXT
	return (bpf_select(dev, rw));
#else /* NeXT */
	return (bpf_select(dev, rw, u.u_procp));
#endif /* NeXT */
}
#endif

/*
 * Support for select() system call
 * Inspired by the code in tty.c for the same purpose.
 *
 * Return true iff the specific operation will not block indefinitely.
 * Otherwise, return false but make a note that a selwakeup() must be done.
 */

#ifdef NeXT

int
bpf_select(dev, rw)
	register dev_t dev;
	int rw;
{
	register struct bpf_d *d;
	register int s;

	if (rw != FREAD)
		return (0);
	/*
	 * An imitation of the FIONREAD ioctl code.
	 */
	d = &bpf_dtab[minor(dev)];

#ifdef NeXT
	bpf_next_log(LOG_DEBUG,
		     "bpf_select: entered (d = 0x%x)\n",
		     (int)d, 0, 0, 0, 0);
#endif

	s = splimp();
#ifdef RING_BUFFER
	if (d->bd_hcount > 0 ||
	    (d->bd_immediate && d->bd_len[d->bd_sidx] != 0)) {
	  /*
	   * There is data waiting.
	   */
	  splx(s);
#ifdef NeXT
	  bpf_next_log(LOG_DEBUG,
		       "bpf_select: data has arrived. leaving ... (d = 0x%x)\n",
		       (int)d, 0, 0, 0, 0);
#endif
	  return (1);
	}
#else /* Not RING_BUFFER */
	if (d->bd_hlen != 0 || (d->bd_immediate && d->bd_slen != 0)) {
		/*
		 * There is data waiting.
		 */
		splx(s);
#ifdef NeXT
		bpf_next_log(LOG_DEBUG,
			     "bpf_select: data has arrived. leaving ... (d = 0x%x)\n",
			     (int)d, 0, 0, 0, 0);
#endif
		return (1);
	}
#endif /* Not RING_BUFFER */
	/*
	 * No data ready.  If there's already a select() waiting on this
	 * minor device then this is a collision.  This shouldn't happen
	 * because minors really should not be shared, but if a process
	 * forks while one of these is open, it is possible that both
	 * processes could select on the same descriptor.
	 */
	if (selthreadcache(&d->bd_selthread)) {
	  /* More than one thread */
	  d->bd_selcoll = 1;
	}

	splx(s);
#ifdef NeXT
	bpf_next_log(LOG_DEBUG,
		     "bpf_select: No data, thread (0x%x), leaviing ...\n",
		     (int)d->bd_selthread, 0, 0, 0, 0);
#endif
	return (0);
}

#else /* NeXT */

int
bpf_select(dev, rw, p)
	register dev_t dev;
	int rw;
	struct proc *p;
{
	register struct bpf_d *d;
	register int s;

	if (rw != FREAD)
		return (0);
	/*
	 * An imitation of the FIONREAD ioctl code.
	 */
	d = &bpf_dtab[minor(dev)];

	s = splimp();
#ifdef RING_BUFFER
	if (d->bcount > 0 || (d->bd_immediate && d->bd_len[d->bd_sidx] != 0)) {
	  /*
	   * There is data waiting.
	   */
	  splx(s);
	  return (1);
	}
#else /* Not RING_BUFFER */
	if (d->bd_hlen != 0 || (d->bd_immediate && d->bd_slen != 0)) {
		/*
		 * There is data waiting.
		 */
		splx(s);
		return (1);
	}
#endif /* Not RING_BUFFER */
#if BSD >= 199103
	selrecord(p, &d->bd_sel);
#else
	/*
	 * No data ready.  If there's already a select() waiting on this
	 * minor device then this is a collision.  This shouldn't happen
	 * because minors really should not be shared, but if a process
	 * forks while one of these is open, it is possible that both
	 * processes could select on the same descriptor.
	 */
	if (d->bd_selproc && d->bd_selproc->p_wchan == (caddr_t)&selwait)
		d->bd_selcoll = 1;
	else
		d->bd_selproc = p;
#endif
	splx(s);
	return (0);
}

#endif /* NeXT */

/*
 * Incoming linkage from device drivers.  Process the packet pkt, of length
 * pktlen, which is stored in a contiguous buffer.  The packet is parsed
 * by each process' filter, and if accepted, stashed into the corresponding
 * buffer.
 */
void
bpf_tap(arg, pkt, pktlen)
	caddr_t arg;
	register u_char *pkt;
	register u_int pktlen;
{
	struct bpf_if *bp;
	register struct bpf_d *d;
	register u_int slen;
	/*
	 * Note that the ipl does not have to be raised at this point.
	 * The only problem that could arise here is that if two different
	 * interfaces shared any data.  This is not the case.
	 */
	bp = (struct bpf_if *)arg;
	for (d = bp->bif_dlist; d != 0; d = d->bd_next) {
		++d->bd_rcount;
		slen = bpf_filter(d->bd_filter, pkt, pktlen, pktlen);
		if (slen != 0)
			catchpacket(d, pkt, pktlen, slen, bcopy);
	}
}

/*
 * Copy data from an mbuf chain into a buffer.  This code is derived
 * from m_copydata in sys/uipc_mbuf.c.
 */
static void
bpf_mcopy(src, dst, len)
	u_char *src;
	u_char *dst;
	register int len;
{
	register struct mbuf *m = (struct mbuf *)src;
	register unsigned count;

	while (len > 0) {
		if (m == 0)
			panic("bpf_mcopy");
		count = MIN(m->m_len, len);
		bcopy(mtod(m, caddr_t), (caddr_t)dst, count);
		m = m->m_next;
		dst += count;
		len -= count;
	}
}

/*
 * Incoming linkage from device drivers, when packet is in an mbuf chain.
 */
void
bpf_mtap(arg, m)
	caddr_t arg;
	struct mbuf *m;
{
	struct bpf_if *bp = (struct bpf_if *)arg;
	struct bpf_d *d;
	u_int pktlen, slen;
	struct mbuf *m0;

	pktlen = 0;
	for (m0 = m; m0 != 0; m0 = m0->m_next)
		pktlen += m0->m_len;

	for (d = bp->bif_dlist; d != 0; d = d->bd_next) {
		++d->bd_rcount;
		slen = bpf_filter(d->bd_filter, (u_char *)m, pktlen, 0);
		if (slen != 0)
			catchpacket(d, (u_char *)m, pktlen, slen, bpf_mcopy);
	}
}

/*
 * Move the packet data from interface memory (pkt) into the
 * store buffer.  Return 1 if it's time to wakeup a listener (buffer full),
 * otherwise 0.  "copy" is the routine called to do the actual data
 * transfer.  bcopy is passed in to copy contiguous chunks, while
 * bpf_mcopy is passed in to copy mbuf chains.  In the latter case,
 * pkt is really an mbuf.
 */
static void
catchpacket(d, pkt, pktlen, snaplen, cpfn)
     register struct bpf_d *d;
     register u_char *pkt;
     register u_int pktlen, snaplen;
     register void (*cpfn)();
{
  register struct bpf_hdr *hp;
  register int totlen, curlen;
  register int hdrlen = d->bd_bif->bif_hdrlen;
#ifdef NeXT
  bpf_next_log(LOG_DEBUG,
	       "catchpacket: entered, d = 0x%x, pktlen = 0x%x, snaplen = 0x%x\n",
	       (int)d, (int)pktlen, (int)snaplen, 0, 0);
#endif
  /*
   * Figure out how many bytes to move.  If the packet is
   * greater or equal to the snapshot length, transfer that
   * much.  Otherwise, transfer the whole packet (unless
   * we hit the buffer size limit).
   */
  totlen = hdrlen + MIN(snaplen, pktlen);
  if (totlen > d->bd_bufsize)
    totlen = d->bd_bufsize;

  /*
   * Round up the end of the previous packet to the next longword.
   */
#ifdef RING_BUFFER
  curlen = BPF_WORDALIGN(d->bd_len[d->bd_sidx]);
  if (curlen + totlen > d->bd_bufsize) {
    /*
     * This packet will overflow the storage buffer.
     * Rotate the buffers if we can, then wakeup any
     * pending reads.
     */
    /* Remark that when the ring buffer system keeps as many packets as
       possible, the "bd_nbuf" elementary buffers are used as
       "bd_nbuf" - 1 hold slots and one store slot. 
       Do not forget make one room for store slot. (S.Adachi, 94/08/10) */
    if (d->bd_hcount == d->bd_nbuf - 1) {
      /*
       * We haven't completed the previous read yet,
       * so drop the packet.
       */
      ++d->bd_dcount;
#ifdef NeXT
      bpf_next_log(LOG_DEBUG,
		   "catchpacket: drop packet!, leaving ...\n",
		   0, 0, 0, 0, 0);
#endif
      return;
    }
    ROTATE_BUFFERS(d);
    bpf_wakeup(d);
    curlen = 0;
  }
#else /* Not RING_BUFFER */
  curlen = BPF_WORDALIGN(d->bd_slen);
  if (curlen + totlen > d->bd_bufsize) {
    /*
     * This packet will overflow the storage buffer.
     * Rotate the buffers if we can, then wakeup any
     * pending reads.
     */
    if (d->bd_fbuf == 0) {
      /*
       * We haven't completed the previous read yet,
       * so drop the packet.
       */
      ++d->bd_dcount;
#ifdef NeXT
      bpf_next_log(LOG_DEBUG,
		   "catchpacket: drop packet!, leaving ...\n",
		   0, 0, 0, 0, 0);
#endif
      return;
    }
    ROTATE_BUFFERS(d);
    bpf_wakeup(d);
    curlen = 0;
  }
#endif /* Not RING_BUFFER */
  else if (d->bd_immediate)
    /*
     * Immediate mode is set.  A packet arrived so any
     * reads should be woken up.
     */
    bpf_wakeup(d);

  /*
   * Append the bpf header.
   */
#ifdef RING_BUFFER
  hp = (struct bpf_hdr *)(d->bd_buf[d->bd_sidx] + curlen);
#else /* Not RING_BUFFER */
  hp = (struct bpf_hdr *)(d->bd_sbuf + curlen);
#endif /* Not RING_BUFFER */
#if BSD >= 199103
  microtime(&hp->bh_tstamp);
#elif defined(sun)
  uniqtime(&hp->bh_tstamp);
#else
  hp->bh_tstamp = time;
#endif
  hp->bh_datalen = pktlen;
  hp->bh_hdrlen = hdrlen;
  /*
   * Copy the packet data into the store buffer and update its length.
   */
  (*cpfn)(pkt, (u_char *)hp + hdrlen, (hp->bh_caplen = totlen - hdrlen));
#ifdef RING_BUFFER
  d->bd_len[d->bd_sidx] = curlen + totlen;
#else /* Not RING_BUFFER */
  d->bd_slen = curlen + totlen;
#endif /* Not RING_BUFFER */
#ifdef NeXT
#ifdef RING_BUFFER
  bpf_next_log(LOG_DEBUG, "catchpacket: packet is stored <%d>. leaving ... (d = 0x%x)\n",
	       (int)d->bd_hcount, (int)d, 0, 0, 0);
#else /* Not RING_BUFFER */
  bpf_next_log(LOG_DEBUG, "catchpacket: packet is stored. leaving ... (d = 0x%x)\n",
	       (int)d, 0, 0, 0, 0);
#endif /* Not RING_BUFFER */
#endif /* NeXT */
}

/*
 * Initialize all nonzero fields of a descriptor.
 */
static int
bpf_allocbufs(d)
	register struct bpf_d *d;
{
#ifdef RING_BUFFER
  d->bd_buf = (caddr_t *)malloc(d->bd_nbuf * sizeof(caddr_t),
				  M_DEVBUF, M_WAITOK);
  if (d->bd_buf == 0)
    return (ENOBUFS);

  d->bd_len = (int *)malloc(d->bd_nbuf * sizeof(int),
			    M_DEVBUF, M_WAITOK);
  if (d->bd_len == 0) {
    free((caddr_t)d->bd_buf, M_DEVBUF);
    return (ENOBUFS);
  }

  {
    int i;
    for (i = 0; i < d->bd_nbuf; i++) {
      d->bd_buf[i] = (caddr_t)malloc(d->bd_bufsize, M_DEVBUF, M_WAITOK);
      if (d->bd_buf[i] == 0) {
	int j;
	for (j = 0; j < i; j++)
	  free((caddr_t)d->bd_buf[j], M_DEVBUF);
	free((caddr_t)d->bd_buf, M_DEVBUF);
	free((caddr_t)d->bd_len, M_DEVBUF);
	return (ENOBUFS);
      }
    }
    for (i = 0; i < d->bd_nbuf; i++)
      d->bd_len[i] = 0;
  }
#else /* Not RING_BUFFER */
  d->bd_fbuf = (caddr_t)malloc(d->bd_bufsize, M_DEVBUF, M_WAITOK);
  if (d->bd_fbuf == 0)
    return (ENOBUFS);

  d->bd_sbuf = (caddr_t)malloc(d->bd_bufsize, M_DEVBUF, M_WAITOK);
  if (d->bd_sbuf == 0) {
    free(d->bd_fbuf, M_DEVBUF);
    return (ENOBUFS);
  }
  d->bd_slen = 0;
  d->bd_hlen = 0;
#endif /* Not RING_BUFFER */
  return (0);
}

/*
 * Free buffers currently in use by a descriptor.
 * Called on close.
 */
static void
bpf_freed(d)
     register struct bpf_d *d;
{
#ifdef NeXT
	bpf_next_log(LOG_DEBUG, "bpf_freed entered (d = 0x%x)\n",
		     (int)d, 0, 0, 0, 0);
#endif
  /*
   * We don't need to lock out interrupts since this descriptor has
   * been detached from its interface and it yet hasn't been marked
   * free.
   */
#ifdef RING_BUFFER
  {
    int i;
    if (d->bd_buf != 0) {
      for (i = 0; i < d->bd_nbuf; i++)
	if (d->bd_buf[i] != 0)
	  free((caddr_t)d->bd_buf[i], M_DEVBUF);
      free((caddr_t)d->bd_buf, M_DEVBUF);
    }
    if (d->bd_len != 0)
      free((caddr_t)d->bd_len, M_DEVBUF);
  }
#else /* Not RING_BUFFER */
  if (d->bd_sbuf != 0) {
    free(d->bd_sbuf, M_DEVBUF);
    if (d->bd_hbuf != 0)
      free(d->bd_hbuf, M_DEVBUF);
    if (d->bd_fbuf != 0)
      free(d->bd_fbuf, M_DEVBUF);
  }
#endif /* Not RING_BUFFER */
  if (d->bd_filter)
    free((caddr_t)d->bd_filter, M_DEVBUF);
  D_MARKFREE(d);
#ifdef NeXT
	bpf_next_log(LOG_DEBUG, "bpf_freed leaving ... (d = 0x%x)\n",
		     (int)d, 0, 0, 0, 0);
#endif
}

/*
 * Attach an interface to bpf.  driverp is a pointer to a (struct bpf_if *)
 * in the driver's softc; dlt is the link layer type; hdrlen is the fixed
 * size of the link header (variable length headers not yet supported).
 */
void
bpfattach(driverp, ifp, dlt, hdrlen)
	caddr_t *driverp;
	struct ifnet *ifp;
	u_int dlt, hdrlen;
{
	struct bpf_if *bp;
	int i;
#if BSD < 199103
	static struct bpf_if bpf_ifs[NBPFILTER];
	static int bpfifno;

#ifdef NeXT
	bpf_next_log(LOG_DEBUG, "bpfattached entered\n", 0, 0, 0, 0, 0);
#endif

	bp = (bpfifno < NBPFILTER) ? &bpf_ifs[bpfifno++] : 0;
#else
	bp = (struct bpf_if *)malloc(sizeof(*bp), M_DEVBUF, M_DONTWAIT);
#endif
	if (bp == 0)
		panic("bpfattach");

	bp->bif_dlist = 0;
	bp->bif_driverp = (struct bpf_if **)driverp;
	bp->bif_ifp = ifp;
	bp->bif_dlt = dlt;

	bp->bif_next = bpf_iflist;
	bpf_iflist = bp;

	*bp->bif_driverp = 0;

	/*
	 * Compute the length of the bpf header.  This is not necessarily
	 * equal to SIZEOF_BPF_HDR because we want to insert spacing such
	 * that the network layer header begins on a longword boundary (for
	 * performance reasons and to alleviate alignment restrictions).
	 */
	bp->bif_hdrlen = BPF_WORDALIGN(hdrlen + SIZEOF_BPF_HDR) - hdrlen;

	/*
	 * Mark all the descriptors free if this hasn't been done.
	 */
	if (!D_ISFREE(&bpf_dtab[0]))
		for (i = 0; i < NBPFILTER; ++i)
			D_MARKFREE(&bpf_dtab[i]);

#ifdef NeXT
	bpf_next_log(LOG_DEBUG, "bpf: %s%d attached\n",
		     (int)ifp->if_name, (int)ifp->if_unit, 0, 0, 0);
#else
	printf("bpf: %s%d attached\n", ifp->if_name, ifp->if_unit);
#endif

#ifdef NeXT
	bpf_next_log(LOG_DEBUG, "bpfattached leaving\n", 0, 0, 0, 0, 0);
#endif
}

#if BSD >= 199103
/* XXX This routine belongs in net/if.c. */
/*
 * Set/clear promiscuous mode on interface ifp based on the truth value
 * of pswitch.  The calls are reference counted so that only the first
 * "on" request actually has an effect, as does the final "off" request.
 * Results are undefined if the "off" and "on" requests are not matched.
 */
int
ifpromisc(ifp, pswitch)
	struct ifnet *ifp;
	int pswitch;
{
	struct ifreq ifr;
	/*
	 * If the device is not configured up, we cannot put it in
	 * promiscuous mode.
	 */
	if ((ifp->if_flags & IFF_UP) == 0)
		return (ENETDOWN);

	if (pswitch) {
		if (ifp->if_pcount++ != 0)
			return (0);
		ifp->if_flags |= IFF_PROMISC;
	} else {
		if (--ifp->if_pcount > 0)
			return (0);
		ifp->if_flags &= ~IFF_PROMISC;
	}
	ifr.ifr_flags = ifp->if_flags;
	return ((*ifp->if_ioctl)(ifp, SIOCSIFFLAGS, (caddr_t)&ifr));
}
#endif

#if BSD < 199103
/*
 * Allocate some memory for bpf.  This is temporary SunOS support, and
 * is admittedly a hack.
 * If resources unavaiable, return 0.
 */
static caddr_t
bpf_alloc(size, canwait)
	register int size;
	register int canwait;
{
	register struct mbuf *m;

#ifdef NeXT
	bpf_next_log(LOG_DEBUG, "bpf_alloc entered\n", 0, 0, 0, 0, 0);
#endif

	if ((unsigned)size > (MCLBYTES-8))
		return 0;

	MGET(m, canwait, MT_DATA);
	if (m == 0)
		return 0;
	if ((unsigned)size > (MLEN-8)) {
		MCLGET(m);
		if (m->m_len != MCLBYTES) {
			m_freem(m);
			return 0;
		}
	}
	*mtod(m, struct mbuf **) = m;
#ifdef NeXT
	bpf_next_log(LOG_DEBUG, "bpf_alloc leaving ... \n", 0, 0, 0, 0, 0);
#endif
	return mtod(m, caddr_t) + 8;
}
#endif
#endif
